from datetime import datetime, date
from typing import Optional

import pandas as pd

from cbrapi.cbr_settings import make_cbr_client
from cbrapi.helpers import normalize_data, guess_date


today = date.today()


def get_metals_prices(
    first_date: Optional[str] = None, last_date: Optional[str] = None, period: str = "D"
) -> pd.DataFrame:
    """
    Get precious metals prices time series from CBR.

    Parameters
    ----------
    first_date : str, optional
        Start date in format 'YYYY-MM-DD'. If not specified, defaults to
        '1999-10-01'.

    last_date : str, optional
        End date in format 'YYYY-MM-DD'. If not specified, defaults to
        current date.

    period: {'D', 'M'}, default 'D'
        Data periodicity. Currently daily ('D') and monthly ('M') frequencies are supported.

    Returns
    -------
    pd.DataFrame
        DataFrame with datetime index and the following columns:
        - GOLD : Gold price (RUB per gram)
        - SILVER : Silver price (RUB per gram)
        - PLATINUM : Platinum price (RUB per gram)
        - PALLADIUM : Palladium price (RUB per gram)

    Notes
    -----
    Prices are provided in Russian Rubles per gram.
    Data is available from October 1999.

    Examples
    --------
    >>> get_metals_prices('2023-01-01', '2023-12-31')
    >>> get_metals_prices(period='M')
    """
    cbr_client = make_cbr_client()
    data1 = guess_date(first_date, default_value="1999-10-01")
    data2 = guess_date(last_date, default_value=str(today))
    metals_xml = cbr_client.service.DragMetDynamic(data1, data2)

    try:
        df = pd.read_xml(metals_xml, xpath=".//DrgMet")
    except ValueError:
        return pd.Series()

    level_1_column_mapping = {
        1: "GOLD",
        2: "SILVER",
        3: "PLATINUM",
        4: "PALLADIUM",
    }

    df = normalize_data(
        data=df, period=period, symbol="DrgMet", level_1=level_1_column_mapping
    )
    return df



from typing import Union
from datetime import datetime, date
import pandas as pd


def pad_missing_periods(
    ts: Union[pd.Series, pd.DataFrame], freq: str = "D"
) -> Union[pd.Series, pd.DataFrame]:
    """
    Pad missing dates and values in the time series.
    """
    name = ts.index.name
    if not isinstance(ts.index, pd.PeriodIndex):
        ts.index = ts.index.to_period(freq)
    ts.sort_index(
        ascending=True, inplace=True
    )  # The order should be ascending to make new Period index
    idx = pd.period_range(start=ts.index[0], end=ts.index[-1], freq=freq)
    ts = ts.reindex(idx, method="pad")
    ts.index.rename(name, inplace=True)
    return ts


def calculate_inverse_rate(close_ts):
    """
    Inverse close values for currency rate data.
    """
    return 1.0 / close_ts


def set_datetime_index(data):
    """
    Set datetime index for DataFrame by detecting date columns.
    """
    if not isinstance(data.index, pd.DatetimeIndex):
        for col in data.columns:
            if any(keyword in str(col) for keyword in ["CDate", "DateMet", "D0", "DT"]):
                if data[col].dtype == "object":
                    data[col] = data[col].str.split("T").str[0]

                data.index = pd.to_datetime(data[col], utc=False)
                data.index.name = "DATE"
                data.drop(col, axis=1, inplace=True)
                break
    return data


def remove_unnecessary_columns(data):
    """
    Remove unnecessary columns from DataFrame.
    """
    data.drop(
        columns=[
            col
            for col in ["id", "rowOrder", "vol", "DateUpdate"]
            if col in data.columns
        ],
        inplace=True,
    )
    return data


def unstack_groups(data, symbol):
    """
    Unstack grouped data based on symbol type.
    """
    if symbol == "DrgMet":
        data = data.groupby([data.index, "CodMet"])["price"].first().unstack()
        data.columns.name = None

    if symbol == "MKR":
        data = (
            data.groupby([data.index, "p1"])[["d1", "d7", "d30", "d90"]]
            .first()
            .unstack(level="p1")
        )
        data.columns = data.columns.rename(None, level=1)
        data.columns.name = None

    return data


def column_rename(data, level_0, level_1):
    """
    Rename columns based on mapping dictionaries.
    """
    if isinstance(data.columns, pd.MultiIndex):
        if level_0:
            new_level_0 = data.columns.levels[0].map(lambda x: level_0.get(str(x), x))
            data.columns = data.columns.set_levels(new_level_0, level=0)

        if level_1:
            new_level_1 = data.columns.levels[1].map(lambda x: level_1.get(str(x), x))
            data.columns = data.columns.set_levels(new_level_1, level=1)

    else:
        if level_1 and isinstance(level_1, dict):
            data = data.rename(columns=level_1)
            available_columns = [col for col in level_1.values() if col in data.columns]
            data = data[available_columns] if available_columns else data

        else:
            if level_1 and not isinstance(level_1, dict):
                if len(data.columns) == 1:
                    data = data.rename(columns={data.columns[0]: str(level_1)})
    return data


def normalize_data(data, period, level_0=None, level_1=None, symbol=None):
    """
    Normalize time series data through multiple processing steps.
    """
    if isinstance(data, pd.Series):
        data = data.to_frame()

    set_datetime_index(data)

    remove_unnecessary_columns(data)

    data = unstack_groups(data, symbol)

    data = column_rename(data, level_0, level_1)

    data = pad_missing_periods(data)

    if period.upper() == "M":
        data = data.resample("M").last()

    if len(data.columns) == 1:
        data = data.squeeze()

    return data


def guess_date(input_date, default_value):
    """
    Create data in datetime format.
    CBR accepts "%Y-%m-%d" format only.
    """
    if input_date:
        try:
            date = datetime.strptime(input_date, "%Y-%m-%d")
        except ValueError:
            date = datetime.strptime(input_date, "%Y-%m")
    else:
        date = datetime.strptime(default_value, "%Y-%m-%d")
    return date


def check_symbol(symbol):
    """
    Check symbol for currency.
    """
    if symbol[:3] not in {s[:3] for s in cur_symbol_set}:
        raise ValueError("This symbol does not exist.")


cur_symbol_set = [
    "AUD.CBR",
    "ATS.CBR",
    "AZN.CBR",
    "DZD.CBR",
    "GBP.CBR",
    "AON.CBR",
    "AMD.CBR",
    "BHD.CBR",
    "BYR.CBR",
    "BYN.CBR",
    "BEF.CBR",
    "BGN.CBR",
    "BOB.CBR",
    "BRL.CBR",
    "HUF.CBR",
    "VND.CBR",
    "HKD.CBR",
    "GRD.CBR",
    "GEL.CBR",
    "DKK.CBR",
    "AED.CBR",
    "USD.CBR",
    "EUR.CBR",
    "EGP.CBR",
    "INR.CBR",
    "IDR.CBR",
    "IRR.CBR",
    "IEP.CBR",
    "ISK.CBR",
    "ESP.CBR",
    "ITL.CBR",
    "KZT.CBR",
    "CAD.CBR",
    "QAR.CBR",
    "KGS.CBR",
    "CNY.CBR",
    "KWD.CBR",
    "CUP.CBR",
    "LVL.CBR",
    "LBP.CBR",
    "LTL.CBR",
    "MDL.CBR",
    "MNT.CBR",
    "DEM.CBR",
    "DEM.CBR",
    "NGN.CBR",
    "NLG.CBR",
    "NZD.CBR",
    "NOK.CBR",
    "OMR.CBR",
    "PLN.CBR",
    "PTE.CBR",
    "SAR.CBR",
    "SAR.CBR",
    "ROL.CBR",
    "RON.CBR",
    "XDR.CBR",
    "SGD.CBR",
    "SRD.CBR",
    "TJS.CBR",
    "THB.CBR",
    "BDT.CBR",
    "TRY.CBR",
    "TMM.CBR",
    "TMT.CBR",
    "UZS.CBR",
    "UAH.CBR",
    "FIM.CBR",
    "FRF.CBR",
    "CZK.CBR",
    "SEK.CBR",
    "CHF.CBR",
    "XEU.CBR",
    "EEK.CBR",
    "ETB.CBR",
    "YUN.CBR",
    "RSD.CBR",
    "ZAR.CBR",
    "KRW.CBR",
    "JPY.CBR",
    "MMK.CBR",
    "ALL.CBR",
    "AOA.CBR",
    "ARS.CBR",
    "AFN.CBR",
    "BWP.CBR",
    "BND.CBR",
    "BIF.CBR",
    "VEF.CBR",
    "KPW.CBR",
    "GMD.CBR",
    "GHS.CBR",
    "GNF.CBR",
    "ZWD.CBR",
    "ZWN.CBR",
    "ZRN.CBR",
    "ZMK.CBR",
    "ILS.CBR",
    "JOD.CBR",
    "IQD.CBR",
    "YER.CBR",
    "KES.CBR",
    "CYP.CBR",
    "COP.CBR",
    "CDF.CBR",
    "CRC.CBR",
    "LAK.CBR",
    "SLL.CBR",
    "LYD.CBR",
    "SZL.CBR",
    "MUR.CBR",
    "MRO.CBR",
    "MKD.CBR",
    "MWK.CBR",
    "MGA.CBR",
    "MYR.CBR",
    "MTL.CBR",
    "MAD.CBR",
    "MXN.CBR",
    "MZN.CBR",
    "NPR.CBR",
    "NIO.CBR",
    "PKR.CBR",
    "PYG.CBR",
    "PEN.CBR",
    "KHR.CBR",
    "SCR.CBR",
    "SYP.CBR",
    "SKK.CBR",
    "SIT.CBR",
    "SOS.CBR",
    "SDG.CBR",
    "TJR.CBR",
    "TWD.CBR",
    "TZS.CBR",
    "TND.CBR",
    "UGX.CBR",
    "UYU.CBR",
    "PHP.CBR",
    "DJF.CBR",
    "XAF.CBR",
    "XOF.CBR",
    "HRK.CBR",
    "CLP.CBR",
    "LKR.CBR",
    "ECS.CBR",
    "NAD.CBR",
]



from importlib.metadata import version


from cbrapi.cbr_settings import make_cbr_client
from cbrapi.currency import get_currencies_list, get_currency_code, get_time_series
from cbrapi.helpers import (
    pad_missing_periods,
    calculate_inverse_rate,
    normalize_data,
    guess_date,
    check_symbol,
)
from cbrapi.ruonia import (
    get_ruonia_ts,
    get_ruonia_index,
    get_ruonia_overnight,
    get_roisfix,
)
from cbrapi.rates import get_key_rate, get_ibor
from cbrapi.metals import get_metals_prices
from cbrapi.reserves import get_mrrf


# __version__ = version("cbr")



from datetime import datetime, date
from typing import Optional

import pandas as pd

from cbrapi.cbr_settings import make_cbr_client
from cbrapi.helpers import normalize_data, guess_date


today = date.today()


def get_mrrf(
    first_date: Optional[str] = None, last_date: Optional[str] = None, period: str = "M"
) -> pd.DataFrame:
    """
    Get International Reserves and Foreign Currency Liquidity data from CBR.

    Parameters
    ----------
    first_date : str, optional
        Start date in format 'YYYY-MM-DD'. If not specified, defaults to
        '1999-01-01'.

    last_date : str, optional
        End date in format 'YYYY-MM-DD'. If not specified, defaults to
        current date.

    period : {'M'}, default 'M'
        Data periodicity. Currently only monthly ('M') frequency is supported.

    Returns
    -------
    pd.DataFrame
        DataFrame with datetime index and the following columns:
        - TOTAL_RESERVES : Total international reserves (USD)
        - CURRENCY_RESERVES : Currency reserves (USD)
        - FOREIGN_CURRENCY : Foreign currency holdings (USD)
        - SDR_ACCOUNT : Special Drawing Rights (SDR) account (USD)
        - IMF_RESERVE : Reserve position in IMF (USD)
        - MONETARY_GOLD : Monetary gold holdings (USD)

    Notes
    -----
    International reserves are external assets that are readily available to and
    controlled by monetary authorities for:
    - Meeting balance of payments financing needs
    - Intervention in exchange markets to affect currency exchange rate
    - Other related purposes

    All values are reported in US dollars.

    Examples
    --------
    >>> get_mrrf('2020-01-01', '2023-12-31')
    >>> get_mrrf(period='M')
    """
    cbr_client = make_cbr_client()
    data1 = guess_date(first_date, default_value="1999-01-01")
    data2 = guess_date(last_date, default_value=str(today))
    mrrf_xml = cbr_client.service.mrrf(data1, data2)

    try:
        df = pd.read_xml(mrrf_xml, xpath=".//mr")
    except ValueError:
        return pd.Series()

    level_1_column_mapping = {
        "p1": "TOTAL_RESERVES",
        "p2": "CURRENCY_RESERVES",
        "p3": "FOREIGN_CURRENCY",
        "p4": "SDR_ACCOUNT",
        "p5": "IMF_RESERVE",
        "p6": "MONETARY_GOLD",
    }

    df = normalize_data(
        data=df, period=period, symbol="mr", level_1=level_1_column_mapping
    )

    return df



from datetime import datetime, date
from typing import Optional

import pandas as pd

from cbrapi.cbr_settings import make_cbr_client
from cbrapi.helpers import normalize_data, guess_date


today = date.today()


def get_ruonia_ts(
    symbol: str,
    first_date: Optional[str] = None,
    last_date: Optional[str] = None,
    period: str = "D",
) -> pd.Series:
    """
    Get RUONIA (Ruble Overnight Index Average) time series data from CBR.

    Parameters
    ----------
    symbol : str
        Financial instrument symbol. Supported symbols:
        - 'RUONIA.INDX' : RUONIA index
        - 'RUONIA_AVG_1M.RATE' : 1-month average rate
        - 'RUONIA_AVG_3M.RATE' : 3-month average rate
        - 'RUONIA_AVG_6M.RATE' : 6-month average rate
        - Other symbols : return overnight RUONIA rates

    first_date : str, optional
        Start date in format 'YYYY-MM-DD'. If not specified, returns
        data from the earliest available date.

    last_date : str, optional
        End date in format 'YYYY-MM-DD'. If not specified, returns
        data up to the most recent available date.

    period: {'D', 'M'}, default 'D'
        Data periodicity. Currently daily ('D') and monthly ('M') frequencies are supported.

    Returns
    -------
    pd.Series
        Time series data for the requested symbol with datetime index.
        Returns empty Series if no data is available for the given parameters.

    Notes
    -----
    Data is sourced from the Central Bank of Russia (CBR) official statistics.
    The function handles API requests and data parsing from CBR web services.

    Examples
    --------
    >>> get_ruonia_ts('RUONIA.INDX', '2023-01-01', '2023-12-31')
    >>> get_ruonia_ts('RUONIA_AVG_3M.RATE')
    """
    cbr_client = make_cbr_client()
    if symbol in [
        "RUONIA.INDX",
        "RUONIA_AVG_1M.RATE",
        "RUONIA_AVG_3M.RATE",
        "RUONIA_AVG_6M.RATE",
    ]:
        ticker = (
            symbol.split(".")[0] if symbol.split(".")[1] == "RATE" else "RUONIA_INDEX"
        )
        df = get_ruonia_index(first_date, last_date).loc[:, ticker]
        if symbol != "RUONIA.INDX":
            df /= 100
        return normalize_data(df, period, symbol)
    else:
        return get_ruonia_overnight(first_date, last_date, period)


def get_ruonia_index(
    first_date: Optional[str] = None, last_date: Optional[str] = None, period: str = "D"
) -> pd.DataFrame:
    """
    Get RUONIA index and averages time series from CBR.

    Parameters
    ----------
    first_date : str, optional
        Start date in format 'YYYY-MM-DD'. If not specified, defaults to
        '2010-01-01'.

    last_date : str, optional
        End date in format 'YYYY-MM-DD'. If not specified, defaults to
        current date.

    period: {'D', 'M'}, default 'D'
        Data periodicity. Currently daily ('D') and monthly ('M') frequencies are supported.

    Returns
    -------
    pd.DataFrame
        DataFrame with datetime index and the following columns:
        - RUONIA_INDEX : RUONIA index value
        - RUONIA_AVG_1M : 1-month average rate (as decimal)
        - RUONIA_AVG_3M : 3-month average rate (as decimal)
        - RUONIA_AVG_6M : 6-month average rate (as decimal)

    Notes
    -----
    RUONIA (Ruble Overnight Index Average) is the weighted average interest rate
    on interbank loans and deposits. It serves as an indicator of the cost of
    unsecured overnight borrowing.

    Examples
    --------
    >>> get_ruonia_index('2023-01-01', '2023-12-31')
    >>> get_ruonia_index(period='D')
    """
    cbr_client = make_cbr_client()
    data1 = guess_date(first_date, default_value="2010-01-01")
    data2 = guess_date(last_date, default_value=str(today))
    ruonia_index_xml = cbr_client.service.RuoniaSV(data1, data2)

    try:
        df = pd.read_xml(ruonia_index_xml, xpath=".//ra")
    except ValueError:
        return pd.Series()

    level_1_column_mapping = {
        "RUONIA_Index": "RUONIA_INDEX",
        "R1W": "RUONIA_AVG_1M",
        "R2W": "RUONIA_AVG_3M",
        "R1M": "RUONIA_AVG_6M",
    }

    df = normalize_data(
        data=df, period=period, symbol="ra", level_1=level_1_column_mapping
    )
    return df


def get_ruonia_overnight(
    first_date: Optional[str] = None, last_date: Optional[str] = None, period: str = "D"
) -> pd.Series:
    """
    Get RUONIA overnight value time series from CBR.

    Parameters
    ----------
    first_date : str, optional
        Start date in format 'YYYY-MM-DD'. If not specified, defaults to
        '2010-01-01'.

    last_date : str, optional
        End date in format 'YYYY-MM-DD'. If not specified, defaults to
        current date.

    period: {'D', 'M'}, default 'D'
        Data periodicity. Currently daily ('D') and monthly ('M') frequencies are supported.

    Returns
    -------
    pd.Series
        Time series of RUONIA overnight rates with datetime index.
        Rates are returned as decimals (e.g., 0.05 for 5%).

    Notes
    -----
    RUONIA (Ruble Overnight Index Average) is the weighted average interest rate
    on interbank loans and deposits. It serves as an indicator of the cost of
    unsecured overnight borrowing.

    Examples
    --------
    >>> get_ruonia_overnight('2023-01-01', '2023-12-31')
    >>> get_ruonia_overnight(period='D')
    """
    cbr_client = make_cbr_client()
    data1 = guess_date(first_date, default_value="2010-01-01")
    data2 = guess_date(last_date, default_value=str(date.today()))
    ruonia_overnight_xml = cbr_client.service.Ruonia(data1, data2)

    try:
        df = pd.read_xml(ruonia_overnight_xml, xpath="//ro")
    except ValueError:
        return pd.Series()

    level_1_column_mapping = {
        "ruo": "RUONIA_OVERNIGHT",
    }

    df = normalize_data(
        data=df, period=period, symbol="ro", level_1=level_1_column_mapping
    )
    df /= 100
    return df


def get_roisfix(
    first_date: Optional[str] = None, last_date: Optional[str] = None, period: str = "D"
) -> pd.DataFrame:
    """
    Get ROISfix (Ruble Overnight Index Swap Fixing) time series from CBR.

    Parameters
    ----------
    first_date : str, optional
        Start date in format 'YYYY-MM-DD'. If not specified, defaults to
        '2011-04-15'.

    last_date : str, optional
        End date in format 'YYYY-MM-DD'. If not specified, defaults to
        current date.

    period: {'D', 'M'}, default 'D'
        Data periodicity. Currently daily ('D') and monthly ('M') frequencies are supported.

    Returns
    -------
    pd.DataFrame
        DataFrame with datetime index and the following columns:
        - RATE_1_WEEK : 1-week ROISfix rate (as decimal)
        - RATE_2_WEEK : 2-week ROISfix rate (as decimal)
        - RATE_1_MONTH : 1-month ROISfix rate (as decimal)
        - RATE_2_MONTH : 2-month ROISfix rate (as decimal)
        - RATE_3_MONTH : 3-month ROISfix rate (as decimal)
        - RATE_6_MONTH : 6-month ROISfix rate (as decimal)

    Notes
    -----
    ROISfix represents the fixed rate in ruble overnight index swaps.

    Examples
    --------
    >>> get_roisfix('2023-01-01', '2023-12-31')
    >>> get_roisfix(period='D')
    """
    cbr_client = make_cbr_client()
    data1 = guess_date(first_date, default_value="2011-04-15")
    data2 = guess_date(last_date, default_value=str(today))
    roisfix_xml = cbr_client.service.ROISfix(data1, data2)

    try:
        df = pd.read_xml(roisfix_xml, xpath=".//rf")
    except ValueError:
        return pd.Series()

    level_1_column_mapping = {
        "R1W": "RATE_1_WEEK",
        "R2W": "RATE_2_WEEK",
        "R1M": "RATE_1_MONTH",
        "R2M": "RATE_2_MONTH",
        "R3M": "RATE_3_MONTH",
        "R6M": "RATE_6_MONTH",
    }

    df = normalize_data(
        data=df, period=period, symbol="rf", level_1=level_1_column_mapping
    )
    return df
